#pike __REAL_VERSION__

//! Compile inline C code to .so and link to it in runtime.
//!
//! Typecal usage is:
//! 
//! object code=C.compile("./inline_code_so","#include<stdio.h>","printf(\"%s\\n\",$(HELLO));return 0;",(["HELLO":""])/*,"gcc -shared" as default*/);
//! int retv=code->eval((["HELLO":"hello world!"]));
//!
//! Simple usage is:
//!
//! int retv=C("#include<stdio.h>","printf(\"%s\\n\",$(HELLO));return 0;",(["HELLO":"hello world!"])/*,"gcc -shared" as default*/);
//! 
//! In typecal usage, compiled code is saved in "./inline_code_so.so". If you keep "./inline_code_so.c" and "./inline_code_so.so" not deleted, following execut will not need to call gcc anymore until your C source code changed.
//!
//! You can pass int or string args in the mapping. At the C side pike-string work as a char*. Return value should be int, you can pass a C-pointer as a int, I think.
//! 
//! Module Dll is needed, and libdl and gcc should have been installed in your system.



//> write("%O",typeof(lambda(int a,int b){}));
//function(int, int : zero)(5) Result: 25

private array function_args_types(function f)
{
	mixed t=_typeof(f);
	string s=sprintf("%O",t);
//werror("typeof : %s\n",s);
	s=s[search(s,"(")+1..];
	s=s[..search(s,":")-1];
	array a=s/",";
	a=map(a,String.trim_all_whites);
	if(a[0]=="")
		a=a[1..];
//werror("%O",a);
return a;
}

private mapping code2dll=([]);

private int __c_inline_sn__;

class _program(string k,mapping args_default){
	int eval(mapping|void args)
	{
		if(args==0)
			args=([]);
		args=args_default|args;
		mapping args_desc=code2dll[k][1];
		array argname_list=indices(args);
		array pos_list=map(indices(args),args_desc);
		sort(pos_list,argname_list);
		return Dll.dllcall(code2dll[k][0]["__pike_c_inline__"],@(map(argname_list,args)));
	}
}

object compile(string so_path,string include,string code,mapping|void args,string|void gcc)
{
	include=include||include_str;
	gcc=gcc||cc_cmdline;
	string k=include+code;
	if(args==0)
		args=([]);
	if(so_path)
		k=so_path;
	if(code2dll[k]==0){
		mapping args_desc=([]);
		string code_out="";
		int last_pos=0;
		int pos;
		pos=search(code,"$(",last_pos);
		int pos2=search(code,")",pos+2);
		int n=1;
		while(pos!=-1&&pos2!=-1){
			code_out+=code[last_pos..pos-1];
			string argname=code[pos+2..pos2-1];
			if(args_desc[argname]==0){
				args_desc[argname]=n;
				n++;
			}
			int m=args_desc[argname];
			if(stringp(args[argname])||zero_type(args[argname]))
				code_out+="((char*)(__pike_c_inline_args__["+(m-1)+"].u.p))";
			else if(intp(args[argname]))
				code_out+="(__pike_c_inline_args__["+(m-1)+"].u.n)";
			else if(floatp(args[argname]))
				code_out+="(__pike_c_inline_args__["+(m-1)+"].u.f)";
			else if(functionp(args[argname])){
				array args_types=function_args_types(args[argname]);
				int k=1;
				foreach(reverse(args_types),string t){
					switch(t){
						case "int":{
							k<<=2;
							break;
						}
						case "string":{
							k<<=2;
							k|=1;
							break;
						}
						case "float":{
							k<<=2;
							k|=2;
							break;
						}
					}
				}
				code_out+="(__T_EXP__("+k+",__pike_c_inline_args__["+(m-1)+"].u.p))";
			}else{
				werror("Unknown type %t.\n",args[argname]);
				return UNDEFINED;
			}

			//code_out+="(__pike_c_inline_args__["+(m-1)+"].t==1?__pike_c_inline_args__["+(m-1)+"].u.n:__pike_c_inline_args__["+(m-1)+"].u.p)";
			last_pos=pos2+1;
			pos=search(code,"$(",last_pos);
			pos2=search(code,")",pos+2);
		}
		code_out+=code[last_pos..];
		string out=include+
#"
#define apply *(long long*)__apply
#define apply_as_string *(char**)__apply
#define apply_as_float *(double*)__apply
struct __A{
	int t;/* 0=none;1=int;2=void*;3=float */
	union U{
		long long n;
		double f;
		void* p;
	}u;
};
struct __T{
	long long t;
	void* p;
};
struct __T __T_EXP__(long long t,void* p)
{
	struct __T r;
	r.t=t;
	r.p=p;
	return r;
};
" 

			+"long long __pike_c_inline__(void* (*__apply)(struct __T p, ...),struct __A* __pike_c_inline_args__)\n{\n"+code_out+"\n}\n";
		if(so_path==0){
			so_path="/tmp/__pike_c_inline_tmp__"+getpid()+"_"+__c_inline_sn__;
		}
		string old=Stdio.read_file(so_path+".c");
		if(old!=out||!Stdio.is_file(so_path+".so")){
			Stdio.write_file(so_path+".c",out);
			if(gcc==0)
				gcc="gcc -g -shared";
			rm(so_path+".so");
//write("cmdline: %s\n",gcc+" "+so_path+".c"+" -o "+so_path+".so");
			Process.system(gcc+" "+so_path+".c"+" -o "+so_path+".so");
		}
//write("so file: %s\n",so_path+".so");

		code2dll[k]=({Dll.dll(so_path+".so"),args_desc});
		__c_inline_sn__++;
	}
	return _program(k,args);
}


int `()(string include,string code,mapping|void args,string|void gcc)
{
	object ob=compile(0,include,code,args,gcc);
	return ob->eval(args);
}


string include_str="";
string cc_cmdline;

void INCLUDE(string s)
{
	include_str=s;
	if(s==0)
		include_str="";
}

void CC(string s)
{
	cc_cmdline=s;
}
