/*
|| This file is part of Pike. For copyright information see COPYRIGHT.
|| Pike is distributed under GPL, LGPL and MPL. See the file COPYING
|| for more information.
|| $Id: dllmod.c,v 0.01 2007/07/05 18:37 peterpan Exp $
*/

#include "global.h"
RCSID("$Id: dllmod.c,v 0.01 2007/07/05 18:37 peterpan Exp $");
#include "dll_machine.h"

#include "interpret.h"
#include "svalue.h"
#include "stralloc.h"
#include "module.h"

#if defined(HAVE_DLFCN_H) && defined(HAVE_LIBDL)
#include <dlfcn.h>

#define sp Pike_sp

struct dll_glue
{
  void * dll;
};

#define THIS ((struct dll_glue *)(Pike_fp->current_storage))

static void do_free(void)
{
  if(THIS->dll)
  {
	dlclose(THIS->dll);
	//XXX: report dlclose error.
  }
}

void dllmod_fatal(char *err)
{
  Pike_error("Dll: %s\n",err);
}

/*! @module Dll
 */

/*! @class dll
 */

/*! @decl void create(string file) //XXX: second arg: void|int flag
 *!
 *! Create dll from a file.
 */

static void dllmod_create(INT32 args)
{
  struct dll_glue *this=THIS;
  do_free();
  if(args)
  {
    void* tmp;
    struct pike_string *tmp2;
    //int rwmode = GDBM_WRCREAT;

    if(sp[-args].type != T_STRING){
      Pike_error("Bad argument 1 to dll->create()\n");
      return;
    }

    /*if(args>1)
    {
      if(sp[1-args].type != T_STRING)
	Pike_error("Bad argument 2 to dll->create()\n");

      rwmode=fixmods(sp[1-args].u.string->str);
    }*/

    tmp2=sp[-args].u.string;

    //THREADS_ALLOW();
    //mt_lock(& gdbm_lock);
    //tmp=dlopen(tmp2->str,RTLD_LAZY);
//printf("tmp2->str=%d\n",(int)(tmp2->str));
//printf("tmp2->str=%s\n",((tmp2->str==NULL)?"NULL":tmp2->str));
    tmp=dlopen(tmp2->str,RTLD_NOW);
//printf("%X\n",tmp);
    //mt_unlock(& gdbm_lock);
    //THREADS_DISALLOW();

    if(!Pike_fp->current_object->prog)
    {
      if(tmp) dlclose(tmp);
      Pike_error("Object destructed in dll->open()n");
    }
    this->dll=tmp;

    pop_n_elems(args);
    if(!this->dll){
	    char buf[1024];
	    sprintf(buf,"Failed to open dll file: %s\n",dlerror());
      Pike_error(buf);
    }
  }
}

/*! @decl int fetch(string key)
 *! @decl int `[](string key)
 *!
 *! Return function address directly, lazy.
 */

static void dllmod_fetch(INT32 args)
{
  struct dll_glue *this=THIS;
  struct pike_string *key;
  void* ret;
  char* err;

  if(!args)
    Pike_error("Too few arguments to dll->fetch()\n");

  if(sp[-args].type != T_STRING)
    Pike_error("Bad argument 1 to dll->fetch()\n");

  if(!THIS->dll)
    Pike_error("Dll not open.\n");

  key=sp[-args].u.string;

  //THREADS_ALLOW();
  //mt_lock(& gdbm_lock);
  dlerror();
//printf("%s\n",key->str);
  ret=dlsym(this->dll, key->str);
  err=dlerror();
  if(err!=NULL)
	printf("%s\n",err);
  //mt_unlock(& gdbm_lock);
  //THREADS_DISALLOW();

  pop_n_elems(args);
/*  if(err)
  {
    push_int(0);//XXX:make distinctions of error and NULL
  }else{*/
//printf("fetch got:%X\n",ret);
    push_int((INT_TYPE)ret);
/*  } */
}




#if 0
struct A{
	int t;/* 0=none;1=int;2=char*;3=array */
	union U{
		int n;
		char* p;
		struct A* a;
	}u;
};

static void dllmod_call_prototype(INT32 args)
{
  (int (*)(int,struct A*)) prototype;
  int f;
  if(args>=2){
	  if(sp[-args].type != T_INT){
		  Pike_error("Bad argument 1 to dll->call_prototype()\n");
	  }else{
		  prototype=(int (*)(int,struct A*))(sp[-args].u.integer);
	  }
	  if(sp[1-args].type != T_INT){
		  Pike_error("Bad argument 2 to dll->call_prototype()\n");
	  }else{
		  f=sp[1-args].u.integer;
	  }
	  struct A *a=malloc(sizeof(struct A)*(args-2));
	  memset(a,0,sizeof(struct A)*(args-2));
	  int error=0;
	  for(int i=2;i<args;i++){
		  switch(sp[i-args].type){
			  case T_INT:
				  a[args-i].t=1;
				  a[args-i].u.n=sp[i-args].u.integer;
				  break;
			  case T_STRING:
				  a[args-i].t=2;
				  a[args-i].u.p=sp[i-args].u.string->str;
				  break;
			  case T_ARRAY:
				  a[args-i].t=3;
				  struct A *p=malloc(sizeof(struct A)*sp[i-args]u.array.size);
				  memset(p,0,sizeof(struct A)*sp[i-args]u.array.size);
				  a[args-1].u.a=p;
				  for(int j=0;j<sp[i-args]u.array.size;j++){
					  struct svalue * item=&(sp[i-args]u.array.item[j]);
					  switch(item->type){
						  case T_INT:
							  p[j].t=1;
							  p[j].u.n=item->u.integer;
							  break;
						  case T_STRING:
							  p[j].t=2;
							  p[j].u.n=item->u.string->str;
							  break;
						  default:
							  error=1;
							  Pike_error("Bad array item %d in argument %d to dll->call_prototype()\n",j+1,i+1);
							  break;

					  }
				  }
				  break;
			  default:
				  error=1;
				  Pike_error("Bad argument %d to dll->call_prototype()\n",i+1);
				  break;


		  }
	  }
	  int ret=0;
	  if(!error&&prototype){
		  ret=prototype(f,a);
	  }
	  if(a){
		  for(int i=0;i<args-2;i++){
			  switch(a[i].type){
				  case 3:
					  free(a[i].u.a);
					  break;
			  }
		  }
		  free(a);
	  }
	  return ret;
  }
}
#endif

/*! @decl void close()
 *!
 *! Closes the Dll.
 */

static void dllmod_close(INT32 args)
{
  pop_n_elems(args);

  do_free();
  push_int(0);
}

static void init_dll_glue(struct object *o)
{
  THIS->dll=0;
}

static void exit_dll_glue(struct object *o)
{
  do_free();
}

/*! @endclass
 */

/*! @decl int dllcall(int f,mixed...args)
 *!
 *! Call f as a function address (!) with args, only int or string args supported, return value should always be int.
 */

struct A{
	int t;/* 0=none;1=int;2=void*;3=float */
	union U{
		long long n;
		double f;
		void* p;
	}u;
};
struct T{
	long long t;
	void* p;
};

union {
	long long res_int;
	double res_float;
	char* res_string;
} res_u;

void* dll_apply(struct T p, ...)
{
	long long t=p.t;
	struct svalue* f=p.p;

	va_list ap;
	long long d;
	double db;
	char *s;
	int args;
	void* res;

	push_svalue(f);
	va_start(ap, p);
	args=0;
	while (t!=1){
		int is_string=t&1;
		int is_float=t&2;
		t>>=2;
		if(is_string){
			s=va_arg(ap, char *);
			//printf("got string: %s\n",s);
			push_string(make_shared_string(s));
		}else if(is_float){
			db=va_arg(ap, double);
			//printf("got string: %s\n",s);
			push_float((FLOAT_TYPE)db);
		}else{
			d=va_arg(ap, long long);
			//printf("got int: %d\n",d);
			push_int((INT_TYPE)d);
		}
		args++;
	}
	va_end(ap);
	f_call_function(args+1);
	//res=(void*)&res_u;
	switch(sp[-1].type){
		case T_INT:
			res_u.res_int=sp[-1].u.integer;
			res=(void*)&res_u.res_int;
			break;
		case T_STRING:
			res_u.res_string=sp[-1].u.string->str;
			res=(void*)&res_u.res_string;
			break;
		case T_FLOAT:
			res_u.res_float=sp[-1].u.float_number;
			res=(void*)&res_u.res_float;
			//printf("float return %f\n",res_u.res_float);
			break;
		default:
			printf("Bad return type: %d\n",sp[-1].type);
			break;
	}
	pop_stack();
	return res;

}


static void dllmod_call(INT32 args)
{
	long long ret=0;
	long long (*f)(void* (*apply)(struct T p, ...),struct A*)=NULL;
	struct A* a;
	int error;
	int i;
	if(args>=1){
		if(sp[-args].type != T_INT){
			Pike_error("Bad argument 1 to dllmod_call()\n");
		}else{
			f=(long long (*)(void* (*apply)(struct T p, ...),struct A*))(sp[-args].u.integer);
		}
	}
	a=malloc(sizeof(struct A)*(args-1));
//printf("malloc...%X\n",a);
	memset(a,0,sizeof(struct A)*(args-1));
	error=0;
	for(i=1;i<args;i++){
		switch(sp[i-args].type){
			case T_INT:
				a[i-1].t=1;
				a[i-1].u.n=sp[i-args].u.integer;
				break;
			case T_FLOAT:
				a[i-1].t=3;
				a[i-1].u.f=sp[i-args].u.float_number;
				break;
			case T_STRING:
				a[i-1].t=2;
				a[i-1].u.p=sp[i-args].u.string->str;
				break;
			case T_FUNCTION:
				a[i-1].t=2;
				a[i-1].u.p=&(sp[i-args]);
				break;
			default:
				error=1;
				Pike_error("Bad argument %d to dllmod_call()\n",i+1);
				break;
		}
	}
	if(!error&&f){
//printf("calling...%X\n",f);
		ret=f(dll_apply,a);
//printf("done...\n");
	}
	if(a){
//printf("free ...%X\n",a);
		free((void*)a);
//printf("done...\n");
	}
	push_int((INT_TYPE)ret);
}

/*! @endmodule
 */


#endif
PIKE_MODULE_EXIT {}

PIKE_MODULE_INIT
{
#if defined(HAVE_DLFCN_H) && defined(HAVE_LIBDL)
  start_new_program();
  ADD_STORAGE(struct dll_glue);
  
  /* function(string) */
  ADD_FUNCTION("create", dllmod_create,
	       tFunc(tStr, tVoid), 0 /*ID_STATIC*/);

  /* function(:void) */
  ADD_FUNCTION("close",dllmod_close,tFunc(tNone,tVoid),0);
  /* function(int:string) */
  ADD_FUNCTION("fetch",dllmod_fetch,tFunc(tStr,tInt),0);
  /* function(int,mixed ... :string) */
  ADD_FUNCTION("`[]",dllmod_fetch,tFunc(tStr,tInt),0);
  set_init_callback(init_dll_glue);
  set_exit_callback(exit_dll_glue);
  end_class("dll",0);
  ADD_FUNCTION("dllcall",dllmod_call,tFunc(tInt tOr(tMixed,tVoid),tInt),0);
  /* function(int:string) */
#endif
}
